#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
import sys
import numbers
import json
from types import ModuleType
try:
    from collections import OrderedDict
except ImportError:
    # simple OrderedDict implementation for Python 2.6
    class OrderedDict(dict):
        def __init__(self, items=(), **kwds):
            items = list(items)
            self._order = [k for k, v in items] + [k for k, v in kwds.items()]
            super(OrderedDict, self).__init__(items)
        def keys(self):
            return self._order
        def values(self):
            return [self[k] for k in self._order]
        def items(self):
            return [(k, self[k]) for k in self._order]
        def __setitem__(self, name, value):
            if name not in self._order:
                self._order.append(name)
            super(OrderedDict, self).__setitem__(name, value)
        def __delitem__(self, name):
            if name in self._order:
                self._order.remove(name)
            super(OrderedDict, self).__delitem__(name)
        def __repr__(self):
            return "OrderedDict([{0}])".format(", ".join("({0}, {1})".format(repr(k), repr(v)) for k, v in self.items()))

import numpy

import oamap.generator
import oamap.compiler
import oamap.extension.common

if sys.version_info[0] > 2:
    basestring = str

# The "PLURTP" type system: Primitives, Lists, Unions, Records, Tuples, and Pointers

class Schema(object):
    _identifier = re.compile("[a-zA-Z][a-zA-Z_0-9]*")   # forbid starting with underscore in field names
    _baddelimiter = re.compile("[a-zA-Z_0-9]")          # could be confused with field names or integers

    def __init__(self, *args, **kwds):
        raise TypeError("Kind cannot be instantiated directly")

    @property
    def nullable(self):
        return self._nullable

    @nullable.setter
    def nullable(self, value):
        if value is not True and value is not False:
            raise TypeError("nullable must be True or False, not {0}".format(repr(value)))
        self._nullable = value

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("mask must be None or an array name (string), not {0}".format(repr(value)))
        self._mask = value

    @property
    def packmask(self):
        return self._packmask

    @packmask.setter
    def packmask(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("packmask must be None or an array name (string), not {0}".format(repr(value)))
        self._packmask = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if value is None:
            self._name = value
            return
        if isinstance(value, basestring):
            match = self._identifier.match(value)
            if match is not None and len(match.group(0)) == len(value):
                self._name = value
                return
        raise TypeError("name must be None or a string matching /{0}/, not {1}".format(repr(value), self._identifier.pattern))

    @property
    def doc(self):
        return self._doc

    @doc.setter
    def doc(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("doc must be None or a string, not {0}".format(repr(value)))
        self._doc = value

    def _labels(self):
        labels = []
        self._collectlabels(set(), labels)
        return labels
        
    def _label(self, labels):
        for index, label in enumerate(labels):
            if label is self:
                return "#{0}".format(index)
        return None

    def __ne__(self, other):
        return not self.__eq__(other)

    def show(self, stream=sys.stdout):
        out = self.__repr__(indent="")
        if stream is None:
            return out
        else:
            stream.write(out)
            stream.write("\n")

    def tojsonfile(self, file, explicit=False, *args, **kwds):
        json.dump(self.tojson(explicit=explicit), file, *args, **kwds)

    def tojsonstring(self, explicit=False, *args, **kwds):
        return json.dumps(self.tojson(explicit=explicit), *args, **kwds)

    def tojson(self, explicit=False):
        return self._tojson(explicit, self._labels(), set())

    @staticmethod
    def fromjsonfile(file, *args, **kwds):
        return Schema.fromjson(json.load(file, *args, **kwds))

    @staticmethod
    def fromjsonstring(data, *args, **kwds):
        return Schema.fromjson(json.loads(data, *args, **kwds))

    @staticmethod
    def fromjson(data):
        labels = {}
        out = Schema._fromjson(data, labels)
        if not isinstance(out, Schema):
            raise TypeError("unresolved label: {0}".format(repr(out)))
        out._finalizefromjson(labels)
        return out

    @staticmethod
    def _fromjson(data, labels):
        if isinstance(data, basestring) and data.startswith("#"):
            return data

        elif isinstance(data, basestring):
            return Primitive._fromjson(data, labels)

        elif isinstance(data, dict):
            tpe = data.get("type", "primitive")
            if tpe == "primitive":
                return Primitive._fromjson(data, labels)
            elif tpe == "list":
                return List._fromjson(data, labels)
            elif tpe == "union":
                return Union._fromjson(data, labels)
            elif tpe == "record":
                return Record._fromjson(data, labels)
            elif tpe == "tuple":
                return Tuple._fromjson(data, labels)
            elif tpe == "pointer":
                return Pointer._fromjson(data, labels)
            else:
                raise TypeError("unrecognized type argument for Schema from JSON: {0}".format(repr(tpe)))

        else:
            raise TypeError("unrecognized type for Schema from JSON: {0}".format(repr(data)))

    def _normalize_extension(self, extension):
        if isinstance(extension, ModuleType):
            recurse = False
            extension = extension.__dict__
        else:
            recurse = True

        if isinstance(extension, dict):
            extension = [extension[n] for n in sorted(extension)]

        try:
            iter(extension)
        except TypeError:
            return []
        else:
            out = []
            for x in extension:
                if isinstance(x, type) and issubclass(x, oamap.generator.ExtendedGenerator):
                    out.append(x)
                elif recurse:
                    out.extend(self._normalize_extension(x))
            return out

    def __call__(self, arrays, prefix="object", delimiter="-", extension=oamap.extension.common):
        return self.generator(prefix=prefix, delimiter=delimiter, extension=self._normalize_extension(extension))(arrays)

    def _finalizegenerator(self, out, cacheidx, memo, extension):
        allgenerators = list(memo.values())
        for generator in memo.values():
            if isinstance(generator, oamap.generator.PointerGenerator):
                # only assign pointer targets after all other types have been resolved
                target, prefix, delimiter = generator.target
                if id(target) in memo:
                    # the target points elsewhere in the type tree: link to that
                    generator._internal = True
                    generator.positions = generator.positions + delimiter + memo[id(target)].derivedname
                    generator.target = memo[id(target)]
                    generator.schema.target = generator.target.schema
                else:
                    # the target is not in the type tree: resolve it now
                    memo2 = OrderedDict()   # new memo, but same cacheidx
                    generator._internal = False
                    generator.target = target._finalizegenerator(target._generator(prefix + delimiter + "X", delimiter, cacheidx, memo2, extension), cacheidx, memo2, extension)
                    generator.schema.target = generator.target.schema
                    for generator2 in memo2.values():
                        allgenerators.append(generator2)

        for generator in allgenerators:
            generator._cachelen = cacheidx[0]

        oamap.generator._uniquestr(out, set())
        return out

################################################################ Primitives can be any Numpy type

class Primitive(Schema):
    def __init__(self, dtype, dims=(), nullable=False, data=None, mask=None, packmask=None, name=None, doc=None):
        self.dtype = dtype
        self.dims = dims
        self.nullable = nullable
        self.data = data
        self.mask = mask
        self.packmask = packmask
        self.name = name
        self.doc = doc

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, value):
        if value is not None and not isinstance(value, numpy.dtype):
            value = numpy.dtype(value)
        self._dtype = value

    @property
    def dims(self):
        return self._dims

    @dims.setter
    def dims(self, value):
        try:
            value = tuple(value)
        except TypeError:
            pass
        if value is not None and (not isinstance(value, tuple) or not all(isinstance(x, numbers.Integral) and x >= 0 for x in value)):
            raise TypeError("dims must be None or a tuple of non-negative integers, not {0}".format(repr(value)))
        self._dims = value

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("data must be None or an array name (string), not {0}".format(repr(value)))
        self._data = value

    def __repr__(self, labels=None, shown=None, indent=None):
        eq = "="

        if labels is None:
            labels = self._labels()
            shown = set()
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))

            args = [repr(self._dtype)]
            if self._dims != ():
                args.append("dims" + eq + repr(self._dims))
            if self._nullable is not False:
                args.append("nullable" + eq + repr(self._nullable))
            if self._data is not None:
                args.append("data" + eq + repr(self._data))
            if self._mask is not None:
                args.append("mask" + eq + repr(self._mask))
            if self._packmask is not None:
                args.append("packmask" + eq + repr(self._packmask))
            if self._name is not None:
                args.append("name" + eq + repr(self._name))

            if label is None:
                return "Primitive(" + ", ".join(args) + ")"
            else:
                return label + ": Primitive(" + ", ".join(args) + ")"

        else:
            return label

    def _collectlabels(self, collection, labels):
        if id(self) not in collection:
            collection.add(id(self))
        else:
            labels.append(self)

    def _tojson(self, explicit, labels, shown):
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))
            if not explicit and self._dtype is not None and self._dims == () and self._nullable is False and self._data is None and self._mask is None and self._packmask is None and self._name is None:
                return str(self._dtype)
            else:
                out = {"type": "primitive", "dtype": None if self._dtype is None else str(self._dtype)}
                if explicit or self._dims != ():
                    out["dims"] = None if self._dims is None else list(self._dims)
                if explicit or self._nullable is not False:
                    out["nullable"] = self._nullable
                if explicit or self._data is not None:
                    out["data"] = self._data
                if explicit or self._mask is not None:
                    out["mask"] = self._mask
                if explicit or self._packmask is not None:
                    out["packmask"] = self._packmask
                if explicit or self._name is not None:
                    out["name"] = self._name
                if explicit or self._doc is not None:
                    out["doc"] = self._doc
                if explicit or label is not None:
                    out["label"] = label
                return out
        else:
            return label

    @staticmethod
    def _fromjson(data, labels):
        if isinstance(data, basestring):
            return Primitive(data)
        else:
            if "dtype" not in data:
                raise TypeError("Primitive Schema from JSON is missing argument 'dtype'")
            out = Primitive(numpy.dtype(data["dtype"]), dims=data.get("dims", []), nullable=data.get("nullable", False), data=data.get("data", None), mask=data.get("mask", None), packmask=data.get("packmask", None), name=data.get("name", None), doc=data.get("doc", None))
            if "label" in data:
                labels[data["label"]] = out
            return out

    def _finalizefromjson(self, labels):
        pass

    def copy(self, **replacements):
        if "dtype" not in replacements:
            replacements["dtype"] = self._dtype
        if "dims" not in replacements:
            replacements["dims"] = self._dims
        if "nullable" not in replacements:
            replacements["nullable"] = self._nullable
        if "data" not in replacements:
            replacements["data"] = self._data
        if "mask" not in replacements:
            replacements["mask"] = self._mask
        if "packmask" not in replacements:
            replacements["packmask"] = self._packmask
        if "name" not in replacements:
            replacements["name"] = self._name
        return Primitive(**replacements)

    def replace(self, fcn, *args, **kwds):
        return fcn(Primitive(self._dtype, dims=self._dims, nullable=self._nullable, data=self._data, mask=self._mask, packmask=self._packmask, name=self._name), *args, **kwds)

    def __eq__(self, other, memo=None):
        return isinstance(other, Primitive) and self.dtype == other.dtype and self.dims == other.dims and self.nullable == other.nullable and self.data == other.data and self.mask == other.mask and self.packmask == other.packmask and self.name == other.name

    def __contains__(self, value, memo=None):
        if self.dtype is None:
            raise TypeError("cannot determine if {0} is in {1}: no dtype specified".format(repr(value), self))
        if self.dims is None:
            raise TypeError("cannot determine if {0} is in {1}: no dims specified".format(repr(value), self))
        if value is None:
            return self.nullable

        def recurse(value, dims):
            if dims == ():
                if issubclass(self.dtype.type, (numpy.bool_, numpy.bool)):
                    return value is True or value is False

                elif issubclass(self.dtype.type, numpy.integer):
                    iinfo = numpy.iinfo(self.dtype.type)
                    return isinstance(value, numbers.Integral) and iinfo.min <= value <= iinfo.max

                elif issubclass(self.dtype.type, numpy.floating):
                    return isinstance(value, numbers.Real)

                elif issubclass(self.dtype.type, numpy.complex):
                    return isinstance(value, numbers.Complex)

                else:
                    raise TypeError("unexpected dtype: {0}".format(self.dtype))

            else:
                try:
                    iter(value)
                    len(value)
                except TypeError:
                    return False
                else:
                    return len(value) == dims[0] and all(recurse(x, dims[1:]) for x in value)

        return recurse(value, self.dims)

    def generator(self, prefix="object", delimiter="-", extension=oamap.extension.common):
        if self._baddelimiter.match(delimiter) is not None:
            raise ValueError("delimiters must not contain /{0}/".self._baddelimiter.pattern)
        cacheidx = [0]
        memo = OrderedDict()
        extension = self._normalize_extension(extension)
        return self._finalizegenerator(self._generator(prefix, delimiter, cacheidx, memo, extension), cacheidx, memo, extension)

    def _generator(self, prefix, delimiter, cacheidx, memo, extension):
        if id(self) in memo:
            raise TypeError("types may not be defined in terms of themselves:\n\n    {0}".format(repr(self)))
        memo[id(self)] = None
        args = []

        if self._name is not None:
            prefix = prefix + delimiter + "N" + self._name

        if self._nullable:
            cls = oamap.generator.MaskedPrimitiveGenerator
            if self._mask is None:
                args.append(prefix + delimiter + "M")
            else:
                args.append(self._mask)
            args.append(cacheidx[0]); cacheidx[0] += 1
            if self._packmask is None:
                args.append(prefix + delimiter + "m")
            else:
                args.append(self._packmask)
        else:
            cls = oamap.generator.PrimitiveGenerator

        if self._data is None:
            args.append(prefix)
        else:
            args.append(self._data)
        args.append(cacheidx[0]); cacheidx[0] += 1

        args.append(self._dtype)
        args.append(self._dims)
        args.append(self._name)
        args.append(prefix)
        args.append(self.copy())

        for ext in extension:
            if ext.matches(self):
                args.insert(0, cls)
                cls = ext
                break

        memo[id(self)] = cls(*args)
        return memo[id(self)]

################################################################ Lists may have arbitrary length

class List(Schema):
    def __init__(self, content, nullable=False, starts=None, stops=None, counts=None, mask=None, packmask=None, name=None, doc=None):
        self.content = content
        self.nullable = nullable
        self.starts = starts
        self.stops = stops
        self.counts = counts
        self.mask = mask
        self.packmask = packmask
        self.name = name
        self.doc = doc

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, value):
        if isinstance(value, basestring):
            value = Primitive(value)
        if not isinstance(value, Schema):
            raise TypeError("content must be a Schema, not {0}".format(repr(value)))
        self._content = value

    @property
    def starts(self):
        return self._starts

    @starts.setter
    def starts(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("starts must be None or an array name (string), not {0}".format(repr(value)))
        self._starts = value

    @property
    def stops(self):
        return self._stops

    @stops.setter
    def stops(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("stops must be None or an array name (string), not {0}".format(repr(value)))
        self._stops = value

    @property
    def counts(self):
        return self._counts

    @counts.setter
    def counts(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("counts must be None or an array name (string), not {0}".format(repr(value)))
        self._counts = value

    def __repr__(self, labels=None, shown=None, indent=None):
        eq = "=" if indent is None else " = "

        if labels is None:
            labels = self._labels()
            shown = set()
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))

            args = []
            if indent is None:
                args.append(self._content.__repr__(labels, shown, indent))
            if self._nullable is not False:
                args.append("nullable" + eq + repr(self._nullable))
            if self._starts is not None:
                args.append("starts" + eq + repr(self._starts))
            if self._stops is not None:
                args.append("stops" + eq + repr(self._stops))
            if self._counts is not None:
                args.append("counts" + eq + repr(self._counts))
            if self._mask is not None:
                args.append("mask" + eq + repr(self._mask))
            if self._packmask is not None:
                args.append("packmask" + eq + repr(self._packmask))
            if self._name is not None:
                args.append("name" + eq + repr(self._name))
            if indent is not None:
                if len(args) > 0:
                    args[0] = "\n" + indent + "  " + args[0]
                args.append("\n" + indent + "  content" + eq + self._content.__repr__(labels, shown, indent + "  ").lstrip() + "\n" + indent)

            if label is None:
                return "List(" + ", ".join(args) + ")"
            else:
                return label + ": List(" + ", ".join(args) + ")"

        else:
            return label

    def _tojson(self, explicit, labels, shown):
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))
            out = {"type": "list", "content": self._content._tojson(explicit, labels, shown)}
            if explicit or self._nullable is not False:
                out["nullable"] = self._nullable
            if explicit or self._starts is not None:
                out["starts"] = self._starts
            if explicit or self._stops is not None:
                out["stops"] = self._stops
            if explicit or self._counts is not None:
                out["counts"] = self._counts
            if explicit or self._mask is not None:
                out["mask"] = self._mask
            if explicit or self._packmask is not None:
                out["packmask"] = self._packmask
            if explicit or self._name is not None:
                out["name"] = self._name
            if explicit or self._doc is not None:
                out["doc"] = self._doc
            if explicit or label is not None:
                out["label"] = label
            return out
        else:
            return label

    @staticmethod
    def _fromjson(data, labels):
        if "content" not in data:
            raise TypeError("List Schema from JSON is missing argument 'content'")
        out = List.__new__(List)
        out._content = Schema._fromjson(data["content"], labels)
        out.nullable = data.get("nullable", False)
        out.starts = data.get("starts", None)
        out.stops = data.get("stops", None)
        out.counts = data.get("counts", None)
        out.mask = data.get("mask", None)
        out.packmask = data.get("packmask", None)
        out.name = data.get("name", None)
        out.doc = data.get("doc", None)
        if "label" in data:
            labels[data["label"]] = out
        return out

    def _finalizefromjson(self, labels):
        if isinstance(self._content, basestring):
            if self._content not in labels:
                raise TypeError("unresolved label: {0}".format(repr(self._content)))
            self._content = labels[self._content]
        else:
            self._content._finalizefromjson(labels)

    def _collectlabels(self, collection, labels):
        if id(self) not in collection:
            collection.add(id(self))
            self._content._collectlabels(collection, labels)
        else:
            labels.append(self)

    def copy(self, **replacements):
        if "content" not in replacements:
            replacements["content"] = self._content
        if "nullable" not in replacements:
            replacements["nullable"] = self._nullable
        if "starts" not in replacements:
            replacements["starts"] = self._starts
        if "stops" not in replacements:
            replacements["stops"] = self._stops
        if "counts" not in replacements:
            replacements["counts"] = self._counts
        if "mask" not in replacements:
            replacements["mask"] = self._mask
        if "packmask" not in replacements:
            replacements["packmask"] = self._packmask
        if "name" not in replacements:
            replacements["name"] = self._name
        return List(**replacements)

    def replace(self, fcn, *args, **kwds):
        return fcn(List(self._content.replace(fcn, *args, **kwds), nullable=self._nullable, starts=self._starts, stops=self._stops, counts=self._counts, mask=self._mask, packmask=self._packmask, name=self._name), *args, **kwds)

    def __eq__(self, other, memo=None):
        if memo is None:
            memo = {}
        if id(self) in memo:
            return memo[id(self)] == id(other)
        if not (isinstance(other, List) and self.starts == other.starts and self.stops == other.stops and self.counts == other.counts and self.mask == other.mask and self.packmask == other.packmask and self.name == other.name):
            return False
        memo[id(self)] = id(other)
        return self.content.__eq__(other.content, memo)

    def __contains__(self, value, memo=None):
        if memo is None:
            memo = {}
        if value is None:
            return self.nullable
        try:
            iter(value)
        except TypeError:
            return False
        else:
            for x in value:
                memo2 = dict(memo) if len(memo) > 0 else memo
                if not self.content.__contains__(x, memo2):
                    return False
            return True

    def generator(self, prefix="object", delimiter="-", extension=oamap.extension.common):
        if self._baddelimiter.match(delimiter) is not None:
            raise ValueError("delimiters must not contain /{0}/".self._baddelimiter.pattern)
        cacheidx = [0]
        memo = OrderedDict()
        extension = self._normalize_extension(extension)
        return self._finalizegenerator(self._generator(prefix, delimiter, cacheidx, memo, extension), cacheidx, memo, extension)

    def _generator(self, prefix, delimiter, cacheidx, memo, extension):
        if id(self) in memo:
            raise TypeError("types may not be defined in terms of themselves:\n\n    {0}".format(repr(self)))
        memo[id(self)] = None
        args = []

        if self._name is not None:
            prefix = prefix + delimiter + "N" + self._name

        if self._nullable:
            cls = oamap.generator.MaskedListGenerator
            if self._mask is None:
                args.append(prefix + delimiter + "M")
            else:
                args.append(self._mask)
            args.append(cacheidx[0]); cacheidx[0] += 1
            if self._packmask is None:
                args.append(prefix + delimiter + "m")
            else:
                args.append(self._packmask)
        else:
            cls = oamap.generator.ListGenerator

        if self._starts is None:
            args.append(prefix + delimiter + "B")
        else:
            args.append(self._starts)
        args.append(cacheidx[0]); cacheidx[0] += 1

        if self._stops is None:
            args.append(prefix + delimiter + "E")
        else:
            args.append(self._stops)
        args.append(cacheidx[0]); cacheidx[0] += 1

        if self._counts is None:
            args.append(prefix + delimiter + "c")
        else:
            args.append(self._counts)

        contentgen = self._content._generator(prefix + delimiter + "L", delimiter, cacheidx, memo, extension)
        args.append(contentgen)
        args.append(self._name)
        args.append(prefix)
        args.append(self.copy(content=contentgen.schema))

        for ext in extension:
            if ext.matches(self):
                args.insert(0, cls)
                cls = ext
                break

        memo[id(self)] = cls(*args)
        return memo[id(self)]

################################################################ Unions may be one of several types

class Union(Schema):
    def __init__(self, possibilities, nullable=False, tags=None, offsets=None, mask=None, packmask=None, name=None, doc=None):
        self.possibilities = possibilities
        self.nullable = nullable
        self.tags = tags
        self.offsets = offsets
        self.mask = mask
        self.packmask = packmask
        self.name = name
        self.doc = doc

    @property
    def possibilities(self):
        return tuple(self._possibilities)

    @possibilities.setter
    def possibilities(self, value):
        self._extend(value, [])

    @property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("tags must be None or an array name (string), not {0}".format(repr(value)))
        self._tags = value

    @property
    def offsets(self):
        return self._offsets

    @offsets.setter
    def offsets(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("offsets must be None or an array name (string), not {0}".format(repr(value)))
        self._offsets = value

    def _extend(self, possibilities, start):
        trial = []
        try:
            for i, x in enumerate(possibilities):
                if isinstance(x, basestring):
                    x = Primitive(x)
                assert isinstance(x, Schema), "possibilities must be an iterable of Schemas; item at {0} is {1}".format(i, repr(x))
                trial.append(x)
        except TypeError:
            raise TypeError("possibilities must be an iterable of Schemas, not {0}".format(repr(possibilities)))
        except AssertionError as err:
            raise TypeError(err.message)
        self._possibilities = start + trial

    def append(self, possibility):
        if isinstance(possibility, basestring):
            possibility = Primitive(possibility)
        if not isinstance(possibility, Schema):
            raise TypeError("possibilities must be Schemas, not {0}".format(repr(possibility)))
        self._possibilities.append(possibility)

    def insert(self, index, possibility):
        if isinstance(possibility, basestring):
            possibility = Primitive(possibility)
        if not isinstance(possibility, Schema):
            raise TypeError("possibilities must be Schemas, not {0}".format(repr(possibility)))
        self._possibilities.insert(index, possibility)

    def extend(self, possibilities):
        self._extend(possibilities, self._possibilities)

    def __getitem__(self, index):
        return self._possibilities[index]

    def __setitem__(self, index, value):
        if isinstance(value, basestring):
            value = Primitive(value)
        if not isinstance(value, Schema):
            raise TypeError("possibilities must be Schemas, not {0}".format(repr(value)))
        self._possibilities[index] = value

    def __repr__(self, labels=None, shown=None, indent=None):
        eq = "=" if indent is None else " = "

        if labels is None:
            labels = self._labels()
            shown = set()
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))

            args = []
            if indent is None:
                args.append("[" + ", ".join(x.__repr__(labels, shown, indent) for x in self._possibilities) + "]")
            if self._nullable is not False:
                args.append("nullable" + eq + repr(self._nullable))
            if self._tags is not None:
                args.append("tags" + eq + repr(self._tags))
            if self._offsets is not None:
                args.append("offsets" + eq + repr(self._offsets))
            if self._mask is not None:
                args.append("mask" + eq + repr(self._mask))
            if self._packmask is not None:
                args.append("packmask" + eq + repr(self._packmask))
            if self._name is not None:
                args.append("name" + eq + repr(self._name))
            if indent is not None:
                if len(args) > 0:
                    args[0] = "\n" + indent + "  " + args[0]
                args.append("\n" + indent + "  possibilities" + eq + "[\n" + indent + "    " + (",\n" + indent + "    ").join(x.__repr__(labels, shown, indent + "    ").lstrip() for x in self._possibilities) + "\n" + indent + "  ]")

            if label is None:
                return "Union(" + ", ".join(args) + ")"
            else:
                return label + ": Union(" + ", ".join(args) + ")"

        else:
            return label

    def _tojson(self, explicit, labels, shown):
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))
            out = {"type": "union", "possibilities": [x._tojson(explicit, labels, shown) for x in self._possibilities]}
            if explicit or self._nullable is not False:
                out["nullable"] = self._nullable
            if explicit or self._tags is not None:
                out["tags"] = self._tags
            if explicit or self._offsets is not None:
                out["offsets"] = self._offsets
            if explicit or self._mask is not None:
                out["mask"] = self._mask
            if explicit or self._packmask is not None:
                out["packmask"] = self._packmask
            if explicit or self._name is not None:
                out["name"] = self._name
            if explicit or self._doc is not None:
                out["doc"] = self._doc
            if explicit or label is not None:
                out["label"] = label
            return out
        else:
            return label

    @staticmethod
    def _fromjson(data, labels):
        if "possibilities" not in data:
            raise TypeError("Union Schema from JSON is missing argument 'possibilities'")
        if not isinstance(data["possibilities"], list):
            raise TypeError("argument 'possibilities' for Union Schema from JSON should be a list, not {0}".format(repr(data["possibilities"])))
        out = Union.__new__(Union)
        out._possibilities = [Schema._fromjson(x, labels) for x in data["possibilities"]]
        out.nullable = data.get("nullable", False)
        out.tags = data.get("tags", None)
        out.offsets = data.get("offsets", None)
        out.mask = data.get("mask", None)
        out.packmask = data.get("packmask", None)
        out.name = data.get("name", None)
        out.doc = data.get("doc", None)
        if "label" in data:
            labels[data["label"]] = out
        return out

    def _finalizefromjson(self, labels):
        for i in range(len(self._possibilities)):
            if isinstance(self._possibilities[i], basestring):
                if self._possibilities[i] not in labels:
                    raise TypeError("unresolved label: {0}".format(repr(self._possibilities[i])))
                self._possibilities[i] = labels[self._possibilities[i]]
            else:
                self._possibilities[i]._finalizefromjson(labels)

    def _collectlabels(self, collection, labels):
        if id(self) not in collection:
            collection.add(id(self))
            for possibility in self._possibilities:
                possibility._collectlabels(collection, labels)
        else:
            labels.append(self)

    def copy(self, **replacements):
        if "possibilities" not in replacements:
            replacements["possibilities"] = self._possibilities
        if "nullable" not in replacements:
            replacements["nullable"] = self._nullable
        if "tags" not in replacements:
            replacements["tags"] = self._tags
        if "offsets" not in replacements:
            replacements["offsets"] = self._offsets
        if "mask" not in replacements:
            replacements["mask"] = self._mask
        if "packmask" not in replacements:
            replacements["packmask"] = self._packmask
        if "name" not in replacements:
            replacements["name"] = self._name
        return Union(**replacements)

    def replace(self, fcn, *args, **kwds):
        return fcn(Union([x.replace(fcn, *args, **kwds) for x in self._possibilities], nullable=self._nullable, tags=self._tags, offsets=self._offsets, mask=self._mask, packmask=self._packmask, name=self._name), *args, **kwds)

    def __eq__(self, other, memo=None):
        if memo is None:
            memo = {}
        if id(self) in memo:
            return memo[id(self)] == id(other)
        if not (isinstance(other, Union) and len(self.possibilities) == len(other.possibilities) and self.nullable == other.nullable and self.tags == other.tags and self.offsets == other.offsets and self.mask == other.mask and self.packmask == other.packmask and self.name == other.name):
            return False
        memo[id(self)] = id(other)
        return all(x.__eq__(y, memo) for x, y in zip(self.possibilities, other.possibilities))

    def __contains__(self, value, memo=None):
        if memo is None:
            memo = {}
        if value is None:
            return self.nullable or any(x.nullable for x in self.possibilities)
        return any(x.__contains__(value, memo) for x in self.possibilities)

    def generator(self, prefix="object", delimiter="-", extension=oamap.extension.common):
        if self._baddelimiter.match(delimiter) is not None:
            raise ValueError("delimiters must not contain /{0}/".self._baddelimiter.pattern)
        cacheidx = [0]
        memo = OrderedDict()
        extension = self._normalize_extension(extension)
        return self._finalizegenerator(self._generator(prefix, delimiter, cacheidx, memo, extension), cacheidx, memo, extension)

    def _generator(self, prefix, delimiter, cacheidx, memo, extension):
        if id(self) in memo:
            raise TypeError("types may not be defined in terms of themselves:\n\n    {0}".format(repr(self)))
        memo[id(self)] = None
        args = []

        if self._name is not None:
            prefix = prefix + delimiter + "N" + self._name

        if self._nullable:
            cls = oamap.generator.MaskedUnionGenerator
            if self._mask is None:
                args.append(prefix + delimiter + "M")
            else:
                args.append(self._mask)
            args.append(cacheidx[0]); cacheidx[0] += 1
            if self._packmask is None:
                args.append(prefix + delimiter + "m")
            else:
                args.append(self._packmask)
        else:
            cls = oamap.generator.UnionGenerator

        if self._tags is None:
            args.append(prefix + delimiter + "T")
        else:
            args.append(self._tags)
        args.append(cacheidx[0]); cacheidx[0] += 1

        if self._offsets is None:
            args.append(prefix + delimiter + "O")
        else:
            args.append(self._offsets)
        args.append(cacheidx[0]); cacheidx[0] += 1

        possibilitiesgen = [x._generator(prefix + delimiter + "U" + repr(i), delimiter, cacheidx, memo, extension) for i, x in enumerate(self._possibilities)]
        args.append(possibilitiesgen)
        args.append(self._name)
        args.append(prefix)
        args.append(self.copy(possibilities=[x.schema for x in possibilitiesgen]))

        for ext in extension:
            if ext.matches(self):
                args.insert(0, cls)
                cls = ext
                break

        memo[id(self)] = cls(*args)
        return memo[id(self)]

################################################################ Records contain fields of known types

class Record(Schema):
    def __init__(self, fields, nullable=False, mask=None, packmask=None, name=None, doc=None):
        self.fields = fields
        self.nullable = nullable
        self.mask = mask
        self.packmask = packmask
        self.name = name
        self.doc = doc

    @property
    def fields(self):
        return dict(self._fields)

    @fields.setter
    def fields(self, value):
        self._extend(value, [])

    def _extend(self, fields, start):
        trial = []
        try:
            for n, x in fields.items():
                assert isinstance(n, basestring), "fields must be a dict from identifier strings to Schemas; the key {0} is not a string".format(repr(n))
                matches = self._identifier.match(n)
                assert matches is not None and len(matches.group(0)) == len(n), "fields must be a dict from identifier strings to Schemas; the key {0} is not an identifier (/{1}/)".format(repr(n), self._identifier.pattern)
                if isinstance(x, basestring):
                    x = Primitive(x)
                assert isinstance(x, Schema), "fields must be a dict from identifier strings to Schemas; the value at key {0} is {1}".format(repr(n), repr(x))
                trial.append((n, x))
        except AttributeError:
            raise TypeError("fields must be a dict from strings to Schemas; {0} is not a dict".format(repr(fields)))
        except AssertionError as err:
            raise TypeError(err.message)
        self._fields = OrderedDict(start + trial)

    def __getitem__(self, index):
        return self._fields[index]

    def __setitem__(self, index, value):
        if isinstance(value, basestring):
            value = Primitive(value)
        if not isinstance(value, Schema):
            raise TypeError("field values must be Schemas, not {0}".format(repr(value)))
        self._fields[index] = value

    def __repr__(self, labels=None, shown=None, indent=None):
        eq = "=" if indent is None else " = "

        if labels is None:
            labels = self._labels()
            shown = set()
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))

            args = []
            if indent is None:
                args.append("{" + ", ".join("{0}: {1}".format(repr(n), x.__repr__(labels, shown, indent)) for n, x in self._fields.items()) + "}")
            if self._nullable is not False:
                args.append("nullable" + eq + repr(self._nullable))
            if self._mask is not None:
                args.append("mask" + eq + repr(self._mask))
            if self._packmask is not None:
                args.append("packmask" + eq + repr(self._packmask))
            if self._name is not None:
                args.append("name" + eq + repr(self._name))
            if indent is not None:
                if len(args) > 0:
                    args[0] = "\n" + indent + "  " + args[0]
                args.append("\n" + indent + "  fields" + eq + "{\n" + indent + "    " + (",\n" + indent + "    ").join("{0}: {1}".format(repr(n), x.__repr__(labels, shown, indent + "    ").lstrip()) for n, x in self._fields.items()) + "\n" + indent + "  }")

            if label is None:
                return "Record(" + ", ".join(args) + ")"
            else:
                return label + ": Record(" + ", ".join(args) + ")"

        else:
            return label

    def _tojson(self, explicit, labels, shown):
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))
            out = {"type": "record", "fields": [[n, x._tojson(explicit, labels, shown)] for n, x in self._fields.items()]}
            if explicit or self._nullable is not False:
                out["nullable"] = self._nullable
            if explicit or self._mask is not None:
                out["mask"] = self._mask
            if explicit or self._packmask is not None:
                out["packmask"] = self._packmask
            if explicit or self._name is not None:
                out["name"] = self._name
            if explicit or self._doc is not None:
                out["doc"] = self._doc
            if explicit or label is not None:
                out["label"] = label
            return out
        else:
            return label

    @staticmethod
    def _fromjson(data, labels):
        if "fields" not in data:
            raise TypeError("Record Schema from JSON is missing argument 'fields'")
        out = Record.__new__(Record)
        if isinstance(data["fields"], list) and all(len(x) == 2 and isinstance(x[0], basestring) for x in data["fields"]):
            out._fields = OrderedDict((n, Schema._fromjson(x, labels)) for n, x in data["fields"])
        elif isinstance(data["fields"], dict) and all(isinstance(x, basestring) for x in data["fields"]):
            out._fields = OrderedDict((n, Schema._fromjson(data["fields"][n], labels)) for n in sorted(data["fields"]))
        else:
            raise TypeError("argument 'fields' for Record Schema from JSON should be a list or dict of key-value pairs (in which the keys are strings), not {0}".format(repr(data["fields"])))
        out.nullable = data.get("nullable", False)
        out.mask = data.get("mask", None)
        out.packmask = data.get("packmask", None)
        out.name = data.get("name", None)
        out.doc = data.get("doc", None)
        if "label" in data:
            labels[data["label"]] = out
        return out

    def _finalizefromjson(self, labels):
        for n in list(self._fields.keys()):
            if isinstance(self._fields[n], basestring):
                if self._fields[n] not in labels:
                    raise TypeError("unresolved label: {0}".format(repr(self._fields[n])))
                self._fields[n] = labels[self._fields[n]]
            else:
                self._fields[n]._finalizefromjson(labels)

    def _collectlabels(self, collection, labels):
        if id(self) not in collection:
            collection.add(id(self))
            for field in self._fields.values():
                field._collectlabels(collection, labels)
        else:
            labels.append(self)

    def copy(self, **replacements):
        if "fields" not in replacements:
            replacements["fields"] = self._fields
        if "nullable" not in replacements:
            replacements["nullable"] = self._nullable
        if "mask" not in replacements:
            replacements["mask"] = self._mask
        if "packmask" not in replacements:
            replacements["packmask"] = self._packmask
        if "name" not in replacements:
            replacements["name"] = self._name
        return Record(**replacements)

    def replace(self, fcn, *args, **kwds):
        return fcn(Record(OrderedDict((n, x.replace(fcn, *args, **kwds)) for n, x in self._fields.items()), nullable=self._nullable, mask=self._mask, packmask=self._packmask, name=self._name), *args, **kwds)

    def __eq__(self, other, memo=None):
        if memo is None:
            memo = {}
        if id(self) in memo:
            return memo[id(self)] == id(other)
        if not (isinstance(other, Record) and set(self._fields) == set(other._fields) and self.nullable == other.nullable and self.mask == other.mask and self.packmask == other.packmask and self.name == other.name):
            return False
        memo[id(self)] = id(other)
        return all(self._fields[n].__eq__(other._fields[n], memo) for n in self._fields)

    def __contains__(self, value, memo=None):
        if memo is None:
            memo = {}
        if value is None:
            return self.nullable
        if isinstance(value, dict):
            return all(n in value and x.__contains__(value[n], memo) for n, x in self.fields.items())
        elif isinstance(value, tuple) and hasattr(value, "_fields"):
            return all(n in value._fields and x.__contains__(getattr(value, n), memo) for n, x in self.fields.items())
        elif isinstance(value, (list, tuple)):
            return False
        else:
            return all(hasattr(value, n) and x.__contains__(getattr(value, n), memo) for n, x in self.fields.items())

    def generator(self, prefix="object", delimiter="-", extension=oamap.extension.common):
        if self._baddelimiter.match(delimiter) is not None:
            raise ValueError("delimiters must not contain /{0}/".self._baddelimiter.pattern)
        cacheidx = [0]
        memo = OrderedDict()
        extension = self._normalize_extension(extension)
        return self._finalizegenerator(self._generator(prefix, delimiter, cacheidx, memo, extension), cacheidx, memo, extension)

    def _generator(self, prefix, delimiter, cacheidx, memo, extension):
        if len(self._fields) == 0:
            raise TypeError("Record has no fields")
        if id(self) in memo:
            raise TypeError("types may not be defined in terms of themselves:\n\n    {0}".format(repr(self)))
        memo[id(self)] = None
        args = []

        if self._name is not None:
            prefix = prefix + delimiter + "N" + self._name

        if self._nullable:
            cls = oamap.generator.MaskedRecordGenerator
            if self._mask is None:
                args.append(prefix + delimiter + "M")
            else:
                args.append(self._mask)
            args.append(cacheidx[0]); cacheidx[0] += 1
            if self._packmask is None:
                args.append(prefix + delimiter + "m")
            else:
                args.append(self._packmask)
        else:
            cls = oamap.generator.RecordGenerator

        fieldsgen = OrderedDict([(n, self._fields[n]._generator(prefix + delimiter + "F" + n, delimiter, cacheidx, memo, extension)) for n in sorted(self._fields)])
        args.append(fieldsgen)
        args.append(self._name)
        args.append(prefix)
        args.append(self.copy(fields=OrderedDict((n, x.schema) for n, x in fieldsgen.items())))

        for ext in extension:
            if ext.matches(self):
                args.insert(0, cls)
                cls = ext
                break

        memo[id(self)] = cls(*args)
        return memo[id(self)]

################################################################ Tuples are like records but with an order instead of field names

class Tuple(Schema):
    def __init__(self, types, nullable=False, mask=None, packmask=None, name=None, doc=None):
        self.types = types
        self.nullable = nullable
        self.mask = mask
        self.packmask = packmask
        self.name = name
        self.doc = doc

    @property
    def types(self):
        return tuple(self._types)

    @types.setter
    def types(self, value):
        self._extend(value, [])

    def _extend(self, types, start):
        trial = []
        try:
            for i, x in enumerate(types):
                if isinstance(x, basestring):
                    x = Primitive(x)
                assert isinstance(x, Schema), "types must be an iterable of Schemas; item at {0} is {1}".format(i, repr(x))
                trial.append(x)
        except TypeError:
            raise TypeError("types must be an iterable of Schemas, not {0}".format(repr(types)))
        except AssertionError as err:
            raise TypeError(err.message)
        self._types = start + trial

    def append(self, item):
        if isinstance(item, basestring):
            item = Primitive(item)
        if not isinstance(item, Schema):
            raise TypeError("types must be Schemas, not {0}".format(repr(item)))
        self._types.append(item)

    def insert(self, index, item):
        if isinstance(item, basestring):
            item = Primitive(item)
        if not isinstance(item, Schema):
            raise TypeError("types must be Schemas, not {0}".format(repr(item)))
        self._types.insert(index, item)

    def extend(self, types):
        self._extend(types, self._types)

    def __getitem__(self, index):
        return self._types[index]

    def __setitem__(self, index, value):
        if isinstance(value, basestring):
            value = Primitive(value)
        if not isinstance(item, Schema):
            raise TypeError("types must be Schemas, not {0}".format(repr(value)))
        self._types[index] = value

    def __repr__(self, labels=None, shown=None, indent=None):
        eq = "=" if indent is None else " = "

        if labels is None:
            labels = self._labels()
            shown = set()
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))

            args = []
            if indent is None:
                args.append("[" + ", ".join(x.__repr__(labels, shown) for x in self._types) + "]")
            if self._nullable is not False:
                args.append("nullable" + eq + repr(self._nullable))
            if self._mask is not None:
                args.append("mask" + eq + repr(self._mask))
            if self._packmask is not None:
                args.append("packmask" + eq + repr(self._packmask))
            if self._name is not None:
                args.append("name" + eq + repr(self._name))
            if indent is not None:
                if len(args) > 0:
                    args[0] = "\n" + indent + "  " + args[0]
                args.append("\n" + indent + "  types" + eq + "[\n" + indent + "    " + (",\n" + indent + "    ").join(x.__repr__(labels, shown, indent + "    ").lstrip() for x in self._types) + "\n" + indent + "  ]")

            if label is None:
                return "Tuple(" + ", ".join(args) + ")"
            else:
                return label + ": Tuple(" + ", ".join(args) + ")"

    def _tojson(self, explicit, labels, shown):
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))
            out = {"type": "tuple", "types": [x._tojson(explicit, labels, shown) for x in self._types]}
            if explicit or self._nullable is not False:
                out["nullable"] = self._nullable
            if explicit or self._mask is not None:
                out["mask"] = self._mask
            if explicit or self._packmask is not None:
                out["packmask"] = self._packmask
            if explicit or self._name is not None:
                out["name"] = self._name
            if explicit or self._doc is not None:
                out["doc"] = self._doc
            if explicit or label is not None:
                out["label"] = label
            return out
        else:
            return label

    @staticmethod
    def _fromjson(data, labels):
        if "types" not in data:
            raise TypeError("Tuple Schema from JSON is missing argument 'types'")
        if not isinstance(data["types"], list):
            raise TypeError("argument 'types' for Tuple Schema from JSON should be a list, not {0}".format(repr(data["types"])))
        out = Tuple.__new__(Tuple)
        out._types = [Schema._fromjson(x, labels) for x in data["types"]]
        out.nullable = data.get("nullable", False)
        out.mask = data.get("mask", None)
        out.packmask = data.get("packmask", None)
        out.name = data.get("name", None)
        out.doc = data.get("doc", None)
        if "label" in data:
            labels[data["label"]] = out
        return out

    def _finalizefromjson(self, labels):
        for i in range(len(self._types)):
            if isinstance(self._types[i], basestring):
                if self._types[i] not in labels:
                    raise TypeError("unresolved label: {0}".format(repr(self._types[i])))
                self._types[i] = labels[self._types[i]]
            else:
                self._types[i]._finalizefromjson(labels)

    def _collectlabels(self, collection, labels):
        if id(self) not in collection:
            collection.add(id(self))
            for item in self._types:
                item._collectlabels(collection, labels)
        else:
            labels.append(self)

    def copy(self, **replacements):
        if "types" not in replacements:
            replacements["types"] = self._types
        if "nullable" not in replacements:
            replacements["nullable"] = self._nullable
        if "mask" not in replacements:
            replacements["mask"] = self._mask
        if "packmask" not in replacements:
            replacements["packmask"] = self._packmask
        if "name" not in replacements:
            replacements["name"] = self._name
        return Tuple(**replacements)

    def replace(self, fcn, *args, **kwds):
        return fcn(Tuple([x.replace(fcn, *args, **kwds) for x in self._types], nullable=self._nullable, mask=self._mask, packmask=self._packmask, name=self._name), *args, **kwds)

    def __eq__(self, other, memo=None):
        if memo is None:
            memo = {}
        if id(self) in memo:
            return memo[id(self)] == id(other)
        if not (isinstance(other, Tuple) and len(self._types) == len(other._types) and self.nullable == other.nullable and self.mask == other.mask and self.packmask == other.packmask and self.name == other.name):
            return False
        memo[id(self)] = id(other)
        return all(x.__eq__(y, memo) for x, y in zip(self._types, other._types))

    def __contains__(self, value, memo=None):
        if memo is None:
            memo = {}
        if value is None:
            return self.nullable
        if isinstance(value, tuple) and len(value) == len(self.types):
            return all(x.__contains__(v, memo) for v, x in zip(value, self.types))
        else:
            return False

    def generator(self, prefix="object", delimiter="-", extension=oamap.extension.common):
        if self._baddelimiter.match(delimiter) is not None:
            raise ValueError("delimiters must not contain /{0}/".self._baddelimiter.pattern)
        cacheidx = [0]
        memo = OrderedDict()
        extension = self._normalize_extension(extension)
        return self._finalizegenerator(self._generator(prefix, delimiter, cacheidx, memo, extension), cacheidx, memo, extension)

    def _generator(self, prefix, delimiter, cacheidx, memo, extension):
        if len(self._types) == 0:
            raise TypeError("Tuple has no types")
        if id(self) in memo:
            raise TypeError("types may not be defined in terms of themselves:\n\n    {0}".format(repr(self)))
        memo[id(self)] = None
        args = []

        if self._name is not None:
            prefix = prefix + delimiter + "N" + self._name

        if self._nullable:
            cls = oamap.generator.MaskedTupleGenerator
            if self._mask is None:
                args.append(prefix + delimiter + "M")
            else:
                args.append(self._mask)
            args.append(cacheidx[0]); cacheidx[0] += 1
            if self._packmask is None:
                args.append(prefix + delimiter + "m")
            else:
                args.append(self._packmask)
        else:
            cls = oamap.generator.TupleGenerator

        typesgen = [x._generator(prefix + delimiter + "F" + repr(i), delimiter, cacheidx, memo, extension) for i, x in enumerate(self._types)]
        args.append(typesgen)
        args.append(self._name)
        args.append(prefix)
        args.append(self.copy(types=[x.schema for x in typesgen]))

        for ext in extension:
            if ext.matches(self):
                args.insert(0, cls)
                cls = ext
                break

        memo[id(self)] = cls(*args)
        return memo[id(self)]

################################################################ Pointers redirect to the contents of other types

class Pointer(Schema):
    def __init__(self, target, nullable=False, positions=None, mask=None, packmask=None, name=None, doc=None):
        self.target = target
        self.nullable = nullable
        self.positions = positions
        self.mask = mask
        self.packmask = packmask
        self.name = name
        self.doc = doc

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, value):
        if isinstance(value, basestring):
            value = Primitive(value)
        if not (value is None or isinstance(value, Schema)):
            raise TypeError("target must be None or a Schema, not {0}".format(repr(value)))
        if value is self:
            raise TypeError("Pointer may not point directly at itself (it would never resolve to a value)")
        self._target = value

    @property
    def positions(self):
        return self._positions

    @positions.setter
    def positions(self, value):
        if not (value is None or isinstance(value, basestring)):
            raise TypeError("positions must be None or an array name (string), not {0}".format(repr(value)))
        self._positions = value

    def __repr__(self, labels=None, shown=None, indent=None):
        eq = "=" if indent is None else " = "

        if labels is None:
            labels = self._labels()
            shown = set()
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))

            args = []
            if indent is None:
                args.append(self._target.__repr__(labels, shown, indent))
            if self._nullable is not False:
                args.append("nullable" + eq + repr(self._nullable))
            if self._positions is not None:
                args.append("positions" + eq + repr(self._positions))
            if self._mask is not None:
                args.append("mask" + eq + repr(self._mask))
            if self._packmask is not None:
                args.append("packmask" + eq + repr(self._packmask))
            if self._name is not None:
                args.append("name" + eq + repr(self._name))
            if indent is not None:
                if len(args) > 0:
                    args[0] = "\n" + indent + "  " + args[0]
                args.append("\n" + indent + "  target" + eq + self._target.__repr__(labels, shown, indent + "  ").lstrip() + "\n" + indent)
                
            if label is None:
                return "Pointer(" + ", ".join(args) + ")"
            else:
                return label + ": Pointer(" + ", ".join(args) + ")"

        else:
            return label

    def _tojson(self, explicit, labels, shown):
        label = self._label(labels)

        if label is None or id(self) not in shown:
            shown.add(id(self))
            out = {"type": "pointer", "target": self._target._tojson(explicit, labels, shown)}
            if explicit or self._nullable is not False:
                out["nullable"] = self._nullable
            if explicit or self._positions is not None:
                out["positions"] = self._positions
            if explicit or self._mask is not None:
                out["mask"] = self._mask
            if explicit or self._packmask is not None:
                out["packmask"] = self._packmask
            if explicit or self._name is not None:
                out["name"] = self._name
            if explicit or self._doc is not None:
                out["doc"] = self._doc
            if explicit or label is not None:
                out["label"] = label
            return out
        else:
            return label

    @staticmethod
    def _fromjson(data, labels):
        if "target" not in data:
            raise TypeError("Pointer Schema from JSON is missing argument 'target'")
        out = Pointer.__new__(Pointer)
        out._target = Schema._fromjson(data["target"], labels)
        out.nullable = data.get("nullable", False)
        out.positions = data.get("positions", None)
        out.mask = data.get("mask", None)
        out.packmask = data.get("packmask", None)
        out.name = data.get("name", None)
        out.doc = data.get("doc", None)
        if "label" in data:
            labels[data["label"]] = out
        return out

    def _finalizefromjson(self, labels):
        if isinstance(self._target, basestring):
            if self._target not in labels:
                raise TypeError("unresolved label: {0}".format(repr(self._target)))
            self._target = labels[self._target]
        else:
            self._target._finalizefromjson(labels)

    def _collectlabels(self, collection, labels):
        if id(self) not in collection:
            collection.add(id(self))
            self._target._collectlabels(collection, labels)
        else:
            labels.append(self)

    def copy(self, **replacements):
        if "target" not in replacements:
            replacements["target"] = self._target
        if "nullable" not in replacements:
            replacements["nullable"] = self._nullable
        if "positions" not in replacements:
            replacements["positions"] = self._positions
        if "mask" not in replacements:
            replacements["mask"] = self._mask
        if "packmask" not in replacements:
            replacements["packmask"] = self._packmask
        if "name" not in replacements:
            replacements["name"] = self._name
        return Pointer(**replacements)

    def replace(self, fcn, *args, **kwds):
        return fcn(Pointer(self._target.replace(fcn, *args, **kwds), nullable=self._nullable, positions=self._positions, mask=self._mask, packmask=self._packmask, name=self._name), *args, **kwds)

    def __eq__(self, other, memo=None):
        if memo is None:
            memo = {}
        if id(self) in memo:
            return memo[id(self)] == id(other)
        if not (isinstance(other, Pointer) and self.nullable == other.nullable and self.positions == other.positions and self.mask == other.mask and self.packmask == other.packmask and self.name == other.name):
            return False
        memo[id(self)] = id(other)
        return self.target.__eq__(other.target, memo)

    def __contains__(self, value, memo=None):
        if memo is None:
            memo = {}
        if id(value) in memo:
            return memo[id(value)] == id(self)
        memo[id(value)] = id(self)
        if value is None:
            return self.nullable
        return self.target.__contains__(value, memo)

    def generator(self, prefix="object", delimiter="-", extension=oamap.extension.common):
        if self._baddelimiter.match(delimiter) is not None:
            raise ValueError("delimiters must not contain /{0}/".self._baddelimiter.pattern)
        cacheidx = [0]
        memo = OrderedDict()
        extension = self._normalize_extension(extension)
        return self._finalizegenerator(self._generator(prefix, delimiter, cacheidx, memo, extension), cacheidx, memo, extension)

    def _generator(self, prefix, delimiter, cacheidx, memo, extension):
        if self._target is None:
            raise TypeError("when creating a Pointer type from a Pointer schema, target must be set to a value other than None")

        memo[id(self)] = None
        args = []

        if self._name is not None:
            prefix = prefix + delimiter + "N" + self._name

        if self._nullable:
            cls = oamap.generator.MaskedPointerGenerator
            if self._mask is None:
                args.append(prefix + delimiter + "M")
            else:
                args.append(self._mask)
            args.append(cacheidx[0]); cacheidx[0] += 1
            if self._packmask is None:
                args.append(prefix + delimiter + "m")
            else:
                args.append(self._packmask)
        else:
            cls = oamap.generator.PointerGenerator

        if self._positions is None:
            args.append(prefix + delimiter + "P")
        else:
            args.append(self._positions)
        args.append(cacheidx[0]); cacheidx[0] += 1

        args.append((self._target, prefix, delimiter))  # placeholder! see _finalizegenerator!
        args.append(self._name)
        args.append(prefix)
        args.append(self.copy())

        for ext in extension:
            if ext.matches(self):
                args.insert(0, cls)
                cls = ext
                break

        memo[id(self)] = cls(*args)
        return memo[id(self)]



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
import json

import oamap.generator
import oamap.inference
import oamap.fillable

def toarrays(fillables):
    return dict((n, x[:]) for n, x in fillables.items())

################################################################ Python data, possibly made by json.load

def fromiterdata(values, generator=None, fillables=None, pointer_fromequal=False):
    for value in values:
        if generator is None:
            generator = List(oamap.inference.fromdata(value)).generator()

        if not isinstance(generator, oamap.generator.Generator):
            generator = generator.generator()

        if fillables is None:
            fillables = oamap.fillable.arrays(generator)

        fromdata(value, generator=generator, fillables=fillables, pointer_fromequal=pointer_fromequal)

    return fillables

def fromdata(value, generator=None, fillables=None, pointer_fromequal=False):
    if generator is None:
        generator = oamap.inference.fromdata(value).generator()

    if not isinstance(generator, oamap.generator.Generator):
        generator = generator.generator()

    if fillables is None:
        fillables = oamap.fillable.arrays(generator)

    # get a list of generators (innermost outward) and make sure they're all starting at the last good entry
    pointers = []
    pointerobjs_keys = []
    targetids_keys = []
    fillables_leaf_to_root = []
    def initialize(gen):
        if isinstance(gen, oamap.generator.PrimitiveGenerator):
            fillables[gen.data].revert()
            forefront = len(fillables[gen.data])
            fillables_leaf_to_root.append(fillables[gen.data])

        elif isinstance(gen, oamap.generator.ListGenerator):
            initialize(gen.content)
            fillables[gen.starts].revert()
            fillables[gen.stops].revert()
            assert len(fillables[gen.starts]) == len(fillables[gen.stops])
            forefront = len(fillables[gen.stops])
            fillables_leaf_to_root.append(fillables[gen.starts])
            fillables_leaf_to_root.append(fillables[gen.stops])

        elif isinstance(gen, oamap.generator.UnionGenerator):
            for x in gen.possibilities:
                initialize(x)
            fillables[gen.tags].revert()
            fillables[gen.offsets].revert()
            assert len(fillables[gen.tags]) == len(fillables[gen.offsets])
            forefront = len(fillables[gen.tags])
            fillables_leaf_to_root.append(fillables[gen.tags])
            fillables_leaf_to_root.append(fillables[gen.offsets])

        elif isinstance(gen, oamap.generator.RecordGenerator):
            uniques = set(initialize(x) for x in gen.fields.values())
            assert len(uniques) == 1
            forefront = list(uniques)[0]

        elif isinstance(gen, oamap.generator.TupleGenerator):
            uniques = set(initialize(x) for x in gen.types)
            assert len(uniques) == 1
            forefront = list(uniques)[0]

        elif isinstance(gen, oamap.generator.PointerGenerator):
            if gen._internal and gen.target is generator and len(fillables[gen.positions]) != 0:
                raise TypeError("the root of a Schema may be the target of a Pointer, but if so, it can only be filled from data once")

            if gen not in pointers:
                pointers.append(gen)
            pointerobjs_keys.append(id(gen))
            targetids_keys.append(id(gen.target))

            if not gen._internal:
                initialize(gen.target)
            fillables[gen.positions].revert()
            forefront = len(fillables[gen.positions])
            fillables_leaf_to_root.append(fillables[gen.positions])

        elif isinstance(gen, oamap.generator.ExtendedGenerator):
            forefront = initialize(gen.generic)

        else:
            raise TypeError("unrecognized generator: {0}".format(repr(gen)))

        if isinstance(gen, oamap.generator.Masked):
            fillables[gen.mask].revert()
            # mask forefront overrides any other arrays
            forefront = len(fillables[gen.mask])
            fillables_leaf_to_root.append(fillables[gen.mask])

        return forefront

    # do the initialize
    initialize(generator)

    # how to get the forefront of various objects (used in 'fill')
    def forefront(gen, secondary=False):
        if not secondary and isinstance(gen, oamap.generator.Masked):
            # mask forefront overrides any other arrays
            return fillables[gen.mask].forefront()

        elif isinstance(gen, oamap.generator.PrimitiveGenerator):
            return fillables[gen.data].forefront()

        elif isinstance(gen, oamap.generator.ListGenerator):
            return fillables[gen.stops].forefront()

        elif isinstance(gen, oamap.generator.UnionGenerator):
            return fillables[gen.tags].forefront()

        elif isinstance(gen, oamap.generator.RecordGenerator):
            for x in gen.fields.values():
                return forefront(x)

        elif isinstance(gen, oamap.generator.TupleGenerator):
            for x in gen.types:
                return forefront(x)

        elif isinstance(gen, oamap.generator.PointerGenerator):
            return len(pointerobjs[id(gen)])

        elif isinstance(gen, oamap.generator.ExtendedGenerator):
            return forefront(gen.generic)

    if forefront(generator) != 0 and not isinstance(generator, ListGenerator):
        raise TypeError("non-Lists can only be filled from data once")

    def unionnullable(union):
        for possibility in union.possibilities:
            if isinstance(possibility, oamap.generator.Masked):
                return True
            elif isinstance(possibility, oamap.generator.UnionGenerator):
                return unionnullable(possibility)
        return False

    # the fill function (recursive)
    pointerat = {}
    def fill(obj, gen, targetids, pointerobjs, at):
        if id(gen) in targetids:
            targetids[id(gen)][id(obj)] = (forefront(gen), obj)

        if obj is None:
            if isinstance(gen, oamap.generator.Masked):
                fillables[gen.mask].append(gen.maskedvalue)
                return   # only mask is filled
            elif isinstance(gen, oamap.generator.UnionGenerator) and unionnullable(gen):
                pass     # mask to fill is in a Union possibility
            elif isinstance(gen, oamap.generator.ExtendedGenerator) and isinstance(gen.generic, oamap.generator.Masked):
                fill(obj, gen.generic, targetids, pointerobjs, at)
                return   # filled the generic generator's mask
            else:
                raise TypeError("cannot fill None where expecting type {0} at {1}".format(gen.schema, at))

        # obj is not None (except for the Union case)
        if isinstance(gen, oamap.generator.Masked):
            fillables[gen.mask].append(forefront(gen, secondary=True))

        if isinstance(gen, oamap.generator.PrimitiveGenerator):
            fillables[gen.data].append(obj)

        elif isinstance(gen, oamap.generator.ListGenerator):
            start = stop = forefront(gen.content)
            try:
                if isinstance(obj, dict) or (isinstance(obj, tuple) and hasattr(obj, "_fields")):
                    raise TypeError
                iter(obj)
            except TypeError:
                raise TypeError("cannot fill {0} where expecting type {1} at {2}".format(repr(obj), gen.schema, at))
            else:
                for x in obj:
                    fill(x, gen.content, targetids, pointerobjs, at + (stop - start,))
                    stop += 1

            fillables[gen.starts].append(start)
            fillables[gen.stops].append(stop)

        elif isinstance(gen, oamap.generator.UnionGenerator):
            tag = None
            for i, possibility in enumerate(gen.possibilities):
                if obj in possibility.schema:
                    tag = i
                    break
            if tag is None:
                raise TypeError("cannot fill {0} where expecting type {1} at {2}".format(repr(obj), gen.schema, at))

            offset = forefront(possibility)
            fill(obj, possibility, targetids, pointerobjs, at + ("tag" + repr(tag),))

            fillables[gen.tags].append(tag)
            fillables[gen.offsets].append(offset)

        elif isinstance(gen, oamap.generator.RecordGenerator):
            if isinstance(obj, dict):
                for n, x in gen.fields.items():
                    if n not in obj:
                        raise TypeError("cannot fill {0} because its {1} field is missing at {2}".format(repr(obj), repr(n), at))
                    fill(obj[n], x, targetids, pointerobjs, at + (n,))
            else:
                for n, x in gen.fields.items():
                    if not hasattr(obj, n):
                        raise TypeError("cannot fill {0} because its {1} field is missing at {2}".format(repr(obj), repr(n), at))
                    fill(getattr(obj, n), x, targetids, pointerobjs, at + (n,))

        elif isinstance(gen, oamap.generator.TupleGenerator):
            for i, x in enumerate(gen.types):
                try:
                    v = obj[i]
                except (TypeError, IndexError):
                    raise TypeError("cannot fill {0} because it does not have a field {1} at {2}".format(repr(obj), i, at))
                else:
                    fill(v, x, targetids, pointerobjs, at + (i,))

        elif isinstance(gen, oamap.generator.PointerGenerator):
            # Pointers will be set after we see all the target values
            pointerobjs[id(gen)].append(obj)
            if id(gen) not in pointerat:
                pointerat[id(gen)] = at

        elif isinstance(gen, oamap.generator.ExtendedGenerator):
            fill(gen.degenerate(obj), gen.generic, targetids, pointerobjs, at)

    # attempt to fill (fillables won't update their 'len' until we 'update')
    targetids = dict((x, {}) for x in targetids_keys)
    pointerobjs = dict((x, []) for x in pointerobjs_keys)
    fill(value, generator, targetids, pointerobjs, ())

    # do the pointers after everything else
    for pointer in pointers:
        while len(pointerobjs[id(pointer)]) > 0:
            pointerobjs2 = {id(pointer): []}
            for obj in pointerobjs[id(pointer)]:
                if id(obj) in targetids[id(pointer.target)] and targetids[id(pointer.target)][id(obj)][1] == obj:
                    # case 1: an object in the target *is* the object in the pointer (same ids)
                    position, _ = targetids[id(pointer.target)][id(obj)]

                else:
                    position = None
                    if pointer_fromequal:
                        # fallback to quadratic complexity search
                        for key, (pos, obj2) in targetids[id(pointer.target)].items():
                            if obj == obj2:
                                position = pos
                                break

                    if position is not None:
                        # case 2: an object in the target *is equal to* the object in the pointer (only check if pointer_fromequal)
                        pass

                    else:
                        # case 3: the object was not found; it must be added to the target (beyond indexes where it can be found)
                        fill(obj, pointer.target, targetids, pointerobjs2, pointerat[id(pointer)])
                        position, _ = targetids[id(pointer.target)][id(obj)]

                # every obj in pointerobjs[id(pointer)] gets *one* append
                fillables[pointer.positions].append(position)

            pointerobjs[id(pointer)] = pointerobjs2[id(pointer)]

    # success! (we're still here)
    for fillable in fillables_leaf_to_root:
        fillable.update()

    # return fillables, which can be evaluated to become arrays
    return fillables

################################################################ helper functions for JSON-derived data and iterables

def fromjson(value, generator=None, fillables=None, pointer_fromequal=False):
    return fromdata(oamap.inference.jsonconventions(value), generator=generator, fillables=fillables, pointer_fromequal=pointer_fromequal)

def fromjsonfile(value, generator=None, fillables=None, pointer_fromequal=False):
    return fromdata(oamap.inference.jsonconventions(json.load(value)), generator=generator, fillables=fillables, pointer_fromequal=pointer_fromequal)

def fromjsonstring(value, generator=None, fillables=None, pointer_fromequal=False):
    return fromdata(oamap.inference.jsonconventions(json.loads(value)), generator=generator, fillables=fillables, pointer_fromequal=pointer_fromequal)

def fromjsonfilestream(values, generator=None, fillables=None, pointer_fromequal=False):
    def iterator():
        j = json.JSONDecoder()
        buf = ""
        while True:
            try:
                obj, i = j.raw_decode(buf)
            except ValueError:
                extra = values.read(8192)
                if len(extra) == 0:
                    break
                else:
                    buf = buf.lstrip() + extra
            else:
                yield oamap.inference.jsonconventions(obj)
                buf = buf[i:].lstrip()

    return fromiterdata(iterator(), generator=generator, fillables=fillables, pointer_fromequal=pointer_fromequal)

def fromjsonstream(values, generator=None, fillables=None, pointer_fromequal=False):
    def iterator():
        j = json.JSONDecoder()
        index = 0
        while True:
            try:
                obj, i = j.raw_decode(values[index:])
            except ValueError:
                break
            yield oamap.inference.jsonconventions(obj)
            _, index = fromjsonstream._pattern.match(values, index + i).span()

    return fromiterdata(iterator(), generator=generator, fillables=fillables, pointer_fromequal=pointer_fromequal)

fromjsonstream._pattern = re.compile("\s*")



#!/usr/bin/env python

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import math
import struct
import sys

import numpy

import oamap.generator

if sys.version_info[0] > 2:
    xrange = range

class Fillable(object):
    def __init__(self, dtype, dims):
        raise NotImplementedError

    def __len__(self):
        return self._len

    def forefront(self):
        return self._chunkindex*self.chunksize + self._indexinchunk

    def append(self, value):
        raise NotImplementedError

    def extend(self, values):
        raise NotImplementedError

    def update(self):
        self._len = self.forefront()

    def revert(self):
        self._chunkindex, self._indexinchunk = divmod(self._len, self.chunksize)

    def close(self):
        pass

    def __getitem__(self, index):
        raise NotImplementedError

    def __array__(self):
        return self[:]

################################################################ make fillables

def _makefillables(generator, fillables, makefillable):
    if isinstance(generator, oamap.generator.Masked):
        fillables[generator.mask] = makefillable(generator.mask, generator.maskdtype, ())

    if isinstance(generator, oamap.generator.PrimitiveGenerator):
        if generator.dtype is None:
            raise ValueError("dtype is unknown (None) for Primitive generator at {0}".format(repr(generator.data)))
        if generator.dims is None:
            raise ValueError("dims is unknown (None) for Primitive generator at {0}".format(repr(generator.data)))
        fillables[generator.data] = makefillable(generator.data, generator.dtype, generator.dims)

    elif isinstance(generator, oamap.generator.ListGenerator):
        fillables[generator.starts] = makefillable(generator.starts, generator.posdtype, ())
        fillables[generator.stops]  = makefillable(generator.stops,  generator.posdtype, ())
        _makefillables(generator.content, fillables, makefillable)

    elif isinstance(generator, oamap.generator.UnionGenerator):
        fillables[generator.tags]    = makefillable(generator.tags,    generator.tagdtype,    ())
        fillables[generator.offsets] = makefillable(generator.offsets, generator.offsetdtype, ())
        for possibility in generator.possibilities:
            _makefillables(possibility, fillables, makefillable)

    elif isinstance(generator, oamap.generator.RecordGenerator):
        for field in generator.fields.values():
            _makefillables(field, fillables, makefillable)

    elif isinstance(generator, oamap.generator.TupleGenerator):
        for field in generator.types:
            _makefillables(field, fillables, makefillable)

    elif isinstance(generator, oamap.generator.PointerGenerator):
        fillables[generator.positions] = makefillable(generator.positions, generator.posdtype, ())
        if not generator._internal:
            _makefillables(generator.target, fillables, makefillable)

    elif isinstance(generator, oamap.generator.ExtendedGenerator):
        _makefillables(generator.generic, fillables, makefillable)

    else:
        raise AssertionError("unrecognized generator type: {0}".format(generator))

def arrays(generator, chunksize=8192):
    if not isinstance(generator, oamap.generator.Generator):
        generator = generator.generator()
    fillables = {}
    _makefillables(generator, fillables, lambda name, dtype, dims: FillableArray(dtype, dims=dims, chunksize=chunksize))
    return fillables

def files(generator, directory, chunksize=8192, lendigits=16):
    if not isinstance(generator, oamap.generator.Generator):
        generator = generator.generator()
    if not os.path.exists(directory):
        os.mkdir(directory)
    fillables = {}
    _makefillables(generator, fillables, lambda name, dtype, dims: FillableFile(os.path.join(directory, name), dtype, dims=dims, chunksize=chunksize, lendigits=lendigits))
    return fillables

def numpyfiles(generator, directory, chunksize=8192, lendigits=16):
    if not isinstance(generator, oamap.generator.Generator):
        generator = generator.generator()
    if not os.path.exists(directory):
        os.mkdir(directory)
    fillables = {}
    _makefillables(generator, fillables, lambda name, dtype, dims: FillableNumpyFile(os.path.join(directory, name), dtype, dims=dims, chunksize=chunksize, lendigits=lendigits))
    return fillables

################################################################ FillableArray

class FillableArray(Fillable):
    # Numpy arrays and list items have 96+8 byte (80+8 byte) overhead in Python 2 (Python 3)
    # compared to 8192 1-byte values (8-byte values), this is 1% overhead (0.1% overhead)
    def __init__(self, dtype, dims=(), chunksize=8192):
        if not isinstance(dtype, numpy.dtype):
            dtype = numpy.dtype(dtype)
        self._data = [numpy.empty((chunksize,) + dims, dtype=dtype)]
        self._len = 0
        self._indexinchunk = 0
        self._chunkindex = 0

    @property
    def dtype(self):
        return self._data[0].dtype

    @property
    def dims(self):
        return self._data[0].shape[1:]

    @property
    def chunksize(self):
        return self._data[0].shape[0]

    def append(self, value):
        if self._indexinchunk >= len(self._data[self._chunkindex]):
            while len(self._data) <= self._chunkindex + 1:
                self._data.append(numpy.empty((self.chunksize,) + self.dims, dtype=self.dtype))
            self._indexinchunk = 0
            self._chunkindex += 1

        self._data[self._chunkindex][self._indexinchunk] = value
        self._indexinchunk += 1

    def extend(self, values):
        chunkindex = self._chunkindex
        indexinchunk = self._indexinchunk

        while len(values) > 0:
            if indexinchunk >= len(self._data[chunkindex]):
                while len(self._data) <= chunkindex + 1:
                    self._data.append(numpy.empty((self.chunksize,) + self.dims, dtype=self.dtype))
                indexinchunk = 0
                chunkindex += 1

            tofill = min(len(values), self.chunksize - indexinchunk)
            self._data[chunkindex][indexinchunk : indexinchunk + tofill] = values[:tofill]
            indexinchunk += tofill
            values = values[tofill:]

        self._chunkindex = chunkindex
        self._indexinchunk = indexinchunk

    def __getitem__(self, index):
        if isinstance(index, slice):
            lenself = len(self)
            step  = 1 if index.step is None else index.step
            if step > 0:
                start = 0       if index.start is None else index.start
                stop  = lenself if index.stop  is None else index.stop
            else:
                start = lenself - 1 if index.start is None else index.start
                stop  = 0           if index.stop  is None else index.stop
                
            if start < 0:
                start += lenself
            if stop < 0:
                stop += lenself

            start = min(lenself, max(0, start))
            stop  = min(lenself, max(0, stop))

            if step == 0:
                raise ValueError("slice step cannot be zero")

            else:
                if step > 0:
                    start_chunkindex = int(math.floor(float(start) / self.chunksize))
                    stop_chunkindex = int(math.ceil(float(stop) / self.chunksize))
                    start_indexinchunk = start - start_chunkindex*self.chunksize
                    stop_indexinchunk = stop - (stop_chunkindex - 1)*self.chunksize
                else:
                    start_chunkindex = int(math.floor(float(start) / self.chunksize))
                    stop_chunkindex = int(math.floor(float(stop) / self.chunksize)) - 1
                    start_indexinchunk = start - start_chunkindex*self.chunksize
                    stop_indexinchunk = stop - (stop_chunkindex + 1)*self.chunksize

                def beginend():
                    offset = 0
                    for chunkindex in xrange(start_chunkindex, stop_chunkindex, 1 if step > 0 else -1):
                        if step > 0:
                            if chunkindex == start_chunkindex:
                                begin = start_indexinchunk
                            else:
                                begin = offset
                            if chunkindex == stop_chunkindex - 1:
                                end = stop_indexinchunk
                            else:
                                end = self.chunksize
                                offset = (begin - self.chunksize) % step
                        else:
                            if chunkindex == start_chunkindex:
                                begin = start_indexinchunk
                            else:
                                begin = self.chunksize - 1 - offset
                            if chunkindex == stop_chunkindex + 1 and index.stop is not None:
                                end = stop_indexinchunk
                            else:
                                end = None
                                offset = (begin - -1) % -step
                        yield chunkindex, begin, end

                length = 0
                for chunkindex, begin, end in beginend():
                    if step > 0:
                        length += int(math.ceil(float(end - begin) / step))
                    elif end is None:
                        length += int(math.ceil(-float(begin + 1) / step))
                    else:
                        length += int(math.ceil(-float(begin - end) / step))

                out = numpy.empty((length,) + self.dims, dtype=self.dtype)
                outi = 0

                for chunkindex, begin, end in beginend():
                    array = self._data[chunkindex][begin:end:step]

                    out[outi : outi + len(array)] = array
                    outi += len(array)
                    if outi >= len(out):
                        break

                return out

        else:
            lenself = len(self)
            normalindex = index if index >= 0 else index + lenself
            if not 0 <= normalindex < lenself:
                raise IndexError("index {0} is out of bounds for size {1}".format(index, lenself))

            chunkindex, indexinchunk = divmod(index, self.chunksize)
            return self._data[chunkindex][indexinchunk]

################################################################ FillableFile

class FillableFile(Fillable):
    def __init__(self, filename, dtype, dims=(), chunksize=8192, lendigits=16):
        if not isinstance(dtype, numpy.dtype):
            dtype = numpy.dtype(dtype)
        self._data = numpy.zeros((chunksize,) + dims, dtype=dtype)  # 'zeros', not 'empty' for security
        self._len = 0
        self._indexinchunk = 0
        self._chunkindex = 0
        self._filename = filename
        self._openfile(filename, lendigits)

    def _openfile(self, filename, lendigits):
        open(filename, "wb", 0).close()
        self._file = open(filename, "r+b", 0)
        self._datapos = 0
        # a plain file has no header

    @property
    def filename(self):
        return self._file.name

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def dims(self):
        return self._data.shape[1:]

    @property
    def chunksize(self):
        return self._data.shape[0]

    def append(self, value):
        self._data[self._indexinchunk] = value
        self._indexinchunk += 1

        if self._indexinchunk == self.chunksize:
            self._flush()
            self._indexinchunk = 0
            self._chunkindex += 1

    def _flush(self):
        self._file.seek(self._datapos + self._chunkindex*self.chunksize*self.dtype.itemsize)
        self._file.write(self._data.tostring())
        
    def extend(self, values):
        chunkindex = self._chunkindex
        indexinchunk = self._indexinchunk

        while len(values) > 0:
            tofill = min(len(values), self.chunksize - indexinchunk)
            self._data[indexinchunk : indexinchunk + tofill] = values[:tofill]
            indexinchunk += tofill
            values = values[tofill:]

            if indexinchunk == self.chunksize:
                self._file.seek(self._datapos + chunkindex*self.chunksize*self.dtype.itemsize)
                self._file.write(self._data.tostring())
                indexinchunk = 0
                chunkindex += 1

        self._chunkindex = chunkindex
        self._indexinchunk = indexinchunk

    def revert(self):
        chunkindex, self._indexinchunk = divmod(self._len, self.chunksize)
        if self._chunkindex != chunkindex:
            self._file.seek(self._datapos + chunkindex*self.chunksize*self.dtype.itemsize)
            olddata = numpy.fromstring(self._file.read(self.chunksize*self.dtype.itemsize), dtype=self.dtype)
            self._data[:len(olddata)] = olddata

        self._chunkindex = chunkindex

    def close(self):
        if hasattr(self, "_file"):
            self._flush()
            self._file.close()

    def __del__(self):
        self.close()

    def __enter__(self, *args, **kwds):
        return self

    def __exit__(self, *args, **kwds):
        self.close()

    def __getitem__(self, value):
        if not self._file.closed:
            self._flush()

        if isinstance(value, slice):
            lenself = len(self)
            if lenself == 0:
                array = numpy.empty((lenself,) + self.dims, dtype=self.dtype)
            else:
                array = numpy.memmap(self.filename, self.dtype, "r", self._datapos, (lenself,) + self.dims, "C")
            if value.start is None and value.stop is None and value.step is None:
                return array
            else:
                return array[value]

        else:
            lenself = len(self)
            normalindex = index if index >= 0 else index + lenself
            if not 0 <= normalindex < lenself:
                raise IndexError("index {0} is out of bounds for size {1}".format(index, lenself))

            if not self._file.closed:
                # since the file's still open, get it from here instead of making a new filehandle
                itemsize = self.dtype.itemsize
                try:
                    self._file.seek(self._datapos + normalindex*itemsize)
                    return numpy.fromstring(self._file.read(itemsize), self.dtype)[0]
                finally:
                    self._file.seek(self._datapos + self._chunkindex*self.chunksize*self.dtype.itemsize)
            else:
                # otherwise, you have to open a new file
                with open(self.filename, "rb") as file:
                    file.seek(self._datapos + normalindex*itemsize)
                    return numpy.fromstring(file.read(itemsize), self.dtype)[0]

################################################################ FillableNumpyFile (FillableFile with a self-describing header)

class FillableNumpyFile(FillableFile):
    def _openfile(self, filename, lendigits):
        magic = b"\x93NUMPY\x01\x00"
        header1 = "{{'descr': {0}, 'fortran_order': False, 'shape': (".format(repr(str(self.dtype))).encode("ascii")
        header2 = "{0}, }}".format(repr((10**lendigits - 1,) + self.dims)).encode("ascii")[1:]

        unpaddedlen = len(magic) + 2 + len(header1) + len(header2)
        paddedlen = int(math.ceil(float(unpaddedlen) / self.dtype.itemsize)) * self.dtype.itemsize
        header2 = header2 + b" " * (paddedlen - unpaddedlen)
        self._lenpos = len(magic) + 2 + len(header1)
        self._datapos = len(magic) + 2 + len(header1) + len(header2)
        assert self._datapos % self.dtype.itemsize == 0

        open(filename, "wb", 0).close()
        self._file = open(filename, "r+b", 0)
        self._formatter = "{0:%dd}" % lendigits
        self._file.write(magic)
        self._file.write(struct.pack("<H", len(header1) + len(header2)))
        self._file.write(header1)
        self._file.write(self._formatter.format(len(self)).encode("ascii"))
        self._file.write(header2[lendigits:])

    def _flush(self):
        super(FillableNumpyFile, self)._flush()
        self._file.seek(self._lenpos)
        self._file.write(self._formatter.format(len(self)).encode("ascii"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2017, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os.path

from setuptools import find_packages
from setuptools import setup

def get_version():
    g = {}
    exec(open(os.path.join("oamap", "version.py")).read(), g)
    return g["__version__"]

setup(name = "oamap",
      version = get_version(),
      packages = find_packages(exclude = ["tests"]),
      scripts = [],
      description = "Toolset for computing directly on hierarchically nested, columnar data, such as Apache Arrow.",
      long_description = """Data analysts are often faced with a choice between speed and flexibility. Tabular data, such as an SQL table or CSV file, can be accessed quickly, which improves the question-and-answer nature of exploratory data analysis. Hierarchically nested data, such as JSON, expresses better the relationship between nested quantities. These relationships *can* be represented with separate, linked tables (i.e. `database normalization <https://en.wikipedia.org/wiki/Database_normalization>`_), but at the cost of complexity for the data analyst and the introduction of expensive joins (see `this question <https://stackoverflow.com/q/38831961/1623645>`_, which got me started on this project). Ideally, we want to perform calculations on JSON-like structures at the speed of SQL.

Tools that analyze tabular data get their performance primarily by laying out data in an intelligent way: computers can access contiguous data more quickly than separated data, whether loading from a disk to memory or from memory to the processor. Datasets with many attributes, of which only a few will be 





OAMap, short for Object-Array Mapping and intended




      
Large datasets can be more compact and faster to access when they are laid out in columns (see `Apache Arrow <https://arrow.apache.org/>`_). Even hierarchically nested data can be presented this way, though converting the data between the columnar form and the object form can degrade performance. Non-hierarchical data (rectangular tables, such as an SQL table) can be accessed faster by not materializing rows (see `Apache Drill <https://drill.apache.org/docs/performance/>`_), but this is more complex for data containing variable-length objects, such as arbitrary-length lists.

OAMap is a suite of tools for performing calculations in this way. The name stands for Object-Array-Map, in analogy with Object-Relational-Mapping (ORM) in relational databases. Pure Python calculations are considerably faster and more memory efficient when datasets are expressed in OAMaps, but the real power comes from *compiling* columized code. This toolset includes `extensions to Numba <http://numba.pydata.org/numba-doc/dev/extending/index.html>`_ that will compile your object-oriented code into native array manipulations. Generally, you'd use uncompiled Python for low latency exploration of the data and Numba-compiled functions for high throughput.

OAMap only strictly depends on Numpy, but `Numba <http://numba.pydata.org/>`_ will accelerate it and `pyarrow <https://arrow.apache.org/docs/python/index.html>`_, `h5py <http://www.h5py.org/>`_, etc. provide hooks for converting data among various formats.""",
      author = "Jim Pivarski (DIANA-HEP)",
      author_email = "pivarski@fnal.gov",
      maintainer = "Jim Pivarski (DIANA-HEP)",
      maintainer_email = "pivarski@fnal.gov",
      url = "https://github.com/diana-hep/oamap",
      download_url = "https://github.com/diana-hep/oamap/releases",
      license = "BSD 3-clause",
      test_suite = "tests",
      install_requires = ["numpy"],
      tests_require = [],
      classifiers = [
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ],
      platforms = "Any",
      )
